# -*- coding: utf-8 -*-
from __future__ import division, print_function
from copy import deepcopy
from os import path, chdir, getcwd
from shutil import rmtree
import collections
import argparse
import sys
import json
import functools
import subprocess
import pipes


class Git(object):
    def __init__(self, verbose=False):
        self.verbose = verbose

    def _call_git(self, *args):
        if self.verbose:
            print("git", " ".join(map(pipes.quote, args)))
        try:
            result = subprocess.check_output(['git'] + list(args), stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            # Make sure we always see what the stdout before re-raising the exception.
            # Otherwise it gets lost.
            print(e.output)
            raise

        if self.verbose:
            print(result)
        return result

    def __getattr__(self, name):
        return functools.partial(self._call_git, name)


class GitTagger(object):
    def __init__(self, directory, verbose=False):
        self.directory = directory
        self.repo_name = None
        self.msg = None
        self.start_dir = None
        self.git = None
        self.verbose = verbose

    @staticmethod
    def get_repo_name(url):
        repo_name = path.basename(url)
        if url.endswith('.git'):
            repo_name = path.splitext(repo_name)[0]
        return repo_name

    def clone(self, repo, branch):
        self.start_dir = getcwd()
        chdir(self.directory)
        self.git = Git(verbose=self.verbose)
        self.git.clone(repo)
        self.repo_name = self.get_repo_name(repo)
        chdir(self.repo_name)
        self.git.checkout(branch)
        chdir(self.start_dir)

    def tag(self, name, revision):
        chdir(path.join(self.directory, self.repo_name))
        try:
            self.git.tag(name, revision)
        except subprocess.CalledProcessError as e:
            # if the tag already exists, we should just carry on
            if "already exists" in e.output:
                print("tag {} already exists, continuing".format(name))
            else:
                raise

        self.msg = 'Tagged {}:{} with {}'.format(self.repo_name, revision, name)
        chdir(self.start_dir)

    def branch(self, name, revision):
        chdir(path.join(self.directory, self.repo_name))
        self.git.branch(name, revision)
        self.msg = 'Branched {}:{} with {}'.format(self.repo_name, revision, name)
        chdir(self.start_dir)

    def push(self, tag_or_branch):
        chdir(path.join(self.directory, self.repo_name))
        self.git.push('origin', tag_or_branch)
        chdir(self.start_dir)

    def clean(self):
        chdir(self.directory)
        rmtree(self.repo_name)
        chdir(self.start_dir)


def branch_tag_repos(directory, name, structure, branch_set=None, push=False, clean=False, verbose=False, tag=True):
    if branch_set is None:
        branch_set = set()
    for repo in structure:
        if repo['type'] != 'Git':
            print("Don't know how to handle material type " + repo['type'])
            print(repo)
            continue
        should_branch = False
        for pipeline in repo['pipelines']:
            if pipeline['name'] in branch_set:
                should_branch = True
        if not (tag or should_branch):
            continue
        desc = repo['description']
        url_part, branch_part = desc.split(',')
        url = url_part.split(':', 1)[1].strip()
        branch = branch_part.split(':', 1)[1].strip()
        rev = repo.get('revision') or repo['tag']
        tagger = GitTagger(directory, verbose)
        tagger.clone(url, branch)
        if should_branch:
            tagger.branch(name, rev)
        else:
            tagger.tag(name, rev)
        if push:
            tagger.push(name)
        if clean:
            tagger.clean()


def check_consistent(structure, label):
    revisions = collections.defaultdict(list)
    pipelines = collections.defaultdict(list)
    for repository in structure:
        description = repository['description']
        revision = repository.get('revision') or repository['tag']
        revisions[description].append(revision)
        for pipeline in repository['pipelines']:
            name = pipeline['name']
            counter = pipeline['counter']
            pipelines[(description, revision)] = "{}/{}".format(name, counter)
    bad = False
    for description, revision_list in revisions.items():
        if len(revision_list) > 1:
            bad = True
            print('Repository {} used with more than one revision:'.format(description))
            for revision in revision_list:
                print('Revision {} used in build {}'.format(revision, pipelines[(description, revision)]))
    if bad:
        raise ValueError('Inconsistencies found in {}'.format(label))


def use_tag_in_repolist(repo_list, tag, repo=None, pipeline=None):
    repo_list_copy = deepcopy(repo_list)
    change_count = 0
    if repo:
        checker = lambda repo, mapping: repo in mapping['description']
        name = repo
    elif pipeline:
        checker = lambda pipeline, mapping: any(pipeline == pl['name'] for pl in mapping['pipelines'])
        name = pipeline
    else:
        raise ValueError('repo or pipeline must be provided')
    for entry in repo_list_copy:
        if checker(name, entry):
            del entry['revision']
            entry['tag'] = tag
            change_count += 1
    return repo_list_copy, change_count


def main_updaterepolist():
    parser = argparse.ArgumentParser(
        description="Update a json file created with gocdrepos to replace the "
                    "revision with a provided tag for a certain repository.")

    parser.add_argument(
        'jsonfile',
        help='Json file as produced by gocdrepos.',
        type=argparse.FileType('r'))
    parser.add_argument(
        '-t', '--tag-name',
        required=True,
        help="Name of branch to create."
    )
    parser.add_argument(
        '-r', '--repository',
        help="Substring of repository path (actually 'Description')."
    )

    pargs = parser.parse_args()
    structure = json.load(pargs.jsonfile)
    json_path = pargs.jsonfile.name
    check_consistent(structure, json_path)
    new_structure, count = use_tag_in_repolist(structure, pargs.tag_name, pargs.repository)
    if count != 1:
        print("Expected one change in file, but noticed", count)
    check_consistent(structure, json_path)
    with open(json_path, 'w') as jsonfile:
        json.dump(new_structure, jsonfile)


def parse_branch_list(fh):
    for line in fh.readlines():
        line = line.strip()
        if line:
            yield line;

def branch_set_from_args(pargs):
    branch_set = set();

    if pargs.branch_list:
        branch_set.update([x.strip() for x in pargs.branch_list.split(',')])

    if pargs.branch_list_from_file:
        branch_set.update(parse_branch_list(pargs.branch_list_from_file))

    return branch_set

def main_branchrepos():
    parser = argparse.ArgumentParser(
        description="Branch a set of Git repositories as provided by json data.")

    parser.add_argument(
        'jsonfile',
        help='Json file as produced by gocdrepos. Read from stdin if no  filename given.',
        nargs='?',
        type=argparse.FileType('r'),
        default=sys.stdin)
    parser.add_argument(
        '-d', '--directory',
        default='/tmp',
        help="Parent directory of repository clones clones. Default to /tmp."
    )
    parser.add_argument(
        '-t', '--tag-name',
        required=True,
        help="Name of branch to create."
    )
    parser.add_argument(
        '-b', '--branch-list',
        help="Comma-separated list of pipeline names. Create branches for these."
    )
    parser.add_argument(
        '-B', '--branch-list-from-file',
        type=argparse.FileType('r'),
        help="Pipeline names read in from file. Create branches for these."
    )
    parser.add_argument(
        '-p', '--push',
        action='store_true',
        help="Push changes to remote repo."
    )
    parser.add_argument(
        '-c', '--clean',
        action='store_true',
        help="Remove cloned repo."
    )
    parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        help="Verbose Git output."
    )

    pargs = parser.parse_args()
    branch_set = branch_set_from_args(pargs)
    structure = json.load(pargs.jsonfile)

    check_consistent(structure, pargs.jsonfile.name)
    branch_tag_repos(
        pargs.directory, pargs.tag_name, structure, branch_set, pargs.push, pargs.clean, pargs.verbose, tag=False
    )


def main():
    parser = argparse.ArgumentParser(
        description="Tag and/or branch a set of Git repositories as provided by json data.")
    parser.add_argument(
        'jsonfile',
        help='Json file as produced by gocdrepos. Read from stdin if no  filename given.',
        nargs='?',
        type=argparse.FileType('r'),
        default=sys.stdin)
    parser.add_argument(
        '-d', '--directory',
        default='/tmp',
        help="Parent directory of repository clones clones. Default to /tmp."
    )
    parser.add_argument(
        '-t', '--tag-name',
        required=True,
        help="Name of tag / branch to create."
    )
    parser.add_argument(
        '-b', '--branch-list',
        help="Comma-separated list of pipeline names. Create branches for these."
    )
    parser.add_argument(
        '-B', '--branch-list-from-file',
        type=argparse.FileType('r'),
        help="Pipeline names read in from file. Create branches for these."
    )
    parser.add_argument(
        '-p', '--push',
        action='store_true',
        help="Push changes to remote repo."
    )
    parser.add_argument(
        '-c', '--clean',
        action='store_true',
        help="Remove cloned repo."
    )
    parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        help="Verbose Git output."
    )

    pargs = parser.parse_args()
    branch_set = branch_set_from_args(pargs)
    structure = json.load(pargs.jsonfile)

    check_consistent(structure, pargs.jsonfile.name)
    branch_tag_repos(pargs.directory, pargs.tag_name, structure, branch_set, pargs.push, pargs.clean, pargs.verbose)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from __future__ import print_function
import sys
import json
import yaml
import requests
from collections import OrderedDict

from goserver_config import CruiseTree


class Goserver(object):
    """
    Manages HTTP communication with the Go server.
    """
    config_xml_rest_path = "/go/admin/restful/configuration/file/{}/xml"

    def __init__(self, config, verbose, config_overrides):
        self.__config = {}
        if config is not None:
            self.__config.update(yaml.load(config))
        self.__config.update(config_overrides)
        self.verbose = verbose
        self._cruise_config_md5 = None
        self.tree = None

    def check_config(self):
        assert 'url' in self.__config, "'url' missing."
        assert not (('username' in self.__config) ^ ('password' in self.__config)), \
            "Need both or neither of ('username', 'password') in configuration."

    def fetch_config(self):
        """
        Fetch configuration from Go server
        """
        self.tree = CruiseTree.fromstring(self.xml_from_url())

    @property
    def __auth(self):
        if 'username' in self.__config:
            return self.__config['username'], self.__config['password']

    @property
    def cruise_xml(self):
        return self.tree.tostring()

    @property
    def cruise_xml_subset(self):
        return self.tree.config_subset_tostring()

    def request(self, action, path, **kwargs):
        action = action.upper()
        url = self.__config['url'] + path
        if self.__auth:
            kwargs['auth'] = self.__auth
        response = requests.request(action, url, **kwargs)
        if response.status_code != 200:
            sys.stderr.write("Failed to {} {}\n".format(action, path))
            sys.stderr.write("status-code: {}\n".format(response.status_code))
            sys.stderr.write(u"text: {}\n".format(response.text))
        return response

    def xml_from_url(self):
        action = 'GET'
        path = self.config_xml_rest_path.format(action)
        response = self.request(action, path)
        if response.status_code != 200:
            raise RuntimeError(str(response.status_code))
        try:
            self._cruise_config_md5 = response.headers['x-cruise-config-md5']
        except KeyError:
            print("Missing 'x-cruise-config-md5' in:", file=sys.stderr)
            print(response.headers, file=sys.stderr)
            raise
        return response.text

    def create_a_pipeline(self, pipeline):
        """
        Add a pipeline to the Go server configuration using the REST API:
        https://api.go.cd/current/#create-a-pipeline

        :param pipeline: Json object as describe in API above.
        """
        path = "/go/api/admin/pipelines"
        data = json.dumps(pipeline)
        headers = {
            'Accept': 'application/vnd.go.cd.v3+json',
            'Content-Type': 'application/json'
        }
        response = self.request('post', path, data=data, headers=headers)
        if response.status_code != 200:
            raise RuntimeError(str(response.status_code))

    def get_pipeline_config(self, pipeline_name):
        path = "/go/api/admin/pipelines/" + pipeline_name
        headers = {
            'Accept': 'application/vnd.go.cd.v3+json'
        }
        response = self.request('get', path, headers=headers)
        if response.status_code != 200:
            raise RuntimeError(str(response.status_code))
        json_data = json.loads(response.text.replace("\\'", "'"),
                               object_pairs_hook=OrderedDict)
        etag = response.headers['etag']
        return etag, json_data

    def edit_pipeline_config(self, pipeline_name, etag, pipeline):
        path = "/go/api/admin/pipelines/" + pipeline_name
        data = json.dumps(pipeline)
        headers = {
            'Accept': 'application/vnd.go.cd.v3+json',
            'Content-Type': 'application/json',
            'If-Match': etag
        }
        response = self.request('put', path, data=data, headers=headers)
        if response.status_code != 200:
            print(response.text)
            raise RuntimeError(str(response.status_code))

    def delete_pipeline_config(self, pipeline_name):
        path = "/go/api/admin/pipelines/" + pipeline_name
        headers = {
            'Accept': 'application/vnd.go.cd.v3+json',
        }
        response = self.request('delete', path, headers=headers)
        if response.status_code != 200:
            print(response.text)
            raise RuntimeError(str(response.status_code))

    def unpause(self, pipeline_name):
        headers = {
            'Confirm': 'true'
        }
        path = "/go/api/pipelines/" + pipeline_name + "/unpause"
        self.request('post', path, headers=headers)

    def get_pipeline_status(self, pipeline_name):
        path = "/go/api/pipelines/" + pipeline_name + "/status"
        headers = {
            'Accept': 'application/json'
        }
        response = self.request('get', path, headers=headers)
        if response.status_code != 200:
            raise RuntimeError(str(response.status_code))
        json_data = json.loads(response.text.replace("\\'", "'"),
                               object_pairs_hook=OrderedDict)
        return json_data

    def get_pipeline_groups(self):
        path = "/go/api/config/pipeline_groups"
        headers = {
            'Accept': 'application/json'
        }
        response = self.request('get', path, headers=headers)
        if response.status_code != 200:
            raise RuntimeError(str(response.status_code))
        json_data = json.loads(response.text.replace("\\'", "'"),
                               object_pairs_hook=OrderedDict)
        return json_data

    def get_pipeline_instance(self, pipeline, instance):
        path = "/go/api/pipelines/" + pipeline + "/instance/" + instance
        headers = {
            'Accept': 'application/json'
        }
        response = self.request('get', path, headers=headers)
        if response.status_code != 200:
            raise RuntimeError(str(response.status_code))
        json_data = json.loads(response.text.replace("\\'", "'"),
                               object_pairs_hook=OrderedDict)
        return json_data

    def patch_environment(
            self,
            env_name,
            pipelines_add=None,
            pipelines_remove=None,
            agents_add=None,
            agents_remove=None):
        path = "/go/api/admin/environments/" + env_name
        pipelines = {}
        if pipelines_add:
            pipelines["add"] = pipelines_add
        if pipelines_remove:
            pipelines["remove"] = pipelines_remove
        agents = {}
        if agents_add:
            agents["add"] = agents_add
        if agents_remove:
            agents["remove"] = agents_remove
        data_structure = {}
        if pipelines:
            data_structure["pipelines"] = pipelines
        if agents:
            data_structure["agents"] = agents
        data = json.dumps(data_structure)
        headers = {
            'Accept': 'application/vnd.go.cd.v1+json',
            'Content-Type': 'application/json'
        }
        response = self.request('patch', path, data=data, headers=headers)
        if response.status_code != 200:
            raise RuntimeError(str(response.status_code))

    def upload_config(self):
        """
        This method pushes a new cruise-config.xml to the go server.
        It's used when there is no REST API for the changes we want to do.
        """
        data = {'xmlFile': self.cruise_xml, 'md5': self._cruise_config_md5}
        action = 'POST'
        response = self.request(action,
                                self.config_xml_rest_path.format(action),
                                data=data)
        if response.status_code != 200:
            sys.stderr.write("status-code: %s\n" % response.status_code)
            # GoCD produces broken JSON???, see
            # https://github.com/gocd/gocd/issues/1472
            json_data = json.loads(response.text.replace("\\'", "'"),
                                   object_pairs_hook=OrderedDict)
            sys.stderr.write("result: %s\n" % json_data["result"])
            sys.stderr.write(
                "originalContent:\n%s\n" % json_data["originalContent"])
            raise RuntimeError(response.status_code)

    def rename_pipeline_group(self, source_name, target_name):
        target_group = None
        for group in self.get_pipeline_groups():
            if group['name'] == target_name:
                target_group = group
                break
        if target_group and target_group["pipelines"]:
            names = ", ".join(p['name'] for p in target_group['pipelines'])
            raise ValueError('Pipeline group "{}" exists and contains pipelines: {}'.format(target_name, names))
        self.fetch_config()
        if target_group:
            self.tree.drop_pipeline_group(target_name)
        self.tree.rename_pipeline_group(source_name, target_name)
        self.upload_config()

    def move_all_pipelines_in_group(self, source_name, target_name):
        target_group = None
        for group in self.get_pipeline_groups():
            if group['name'] == target_name:
                target_group = group
                break
        if target_group and target_group["pipelines"]:
            names = ", ".join(p['name'] for p in target_group['pipelines'])
            raise ValueError('Pipeline group "{}" exists and contains pipelines: {}'.format(target_name, names))
        self.fetch_config()
        if target_group:
            self.tree.drop_pipeline_group(target_name)
        self.tree.move_all_pipelines_in_group(source_name, target_name)
        self.upload_config()


